import base64
import struct
import zlib
import xml.etree.ElementTree as ET
from xml.dom import minidom
import numpy as np

def write_mzml(filename, spectra_list):
    """
    Writes a list of spectra to an mzML file.
    
    spectra_list: list of dicts, each containing:
      'mz': numpy array of m/z
      'intensity': numpy array of intensity
      'rt': retention time in seconds
      'id': scan id (optional)
      'ms_level': ms level (default 1)
    """
    
    mzml = ET.Element('mzML', xmlns="http://psi.hupo.org/ms/mzml", version="1.1.0")
    cvList = ET.SubElement(mzml, 'cvList', count="2")
    ET.SubElement(cvList, 'cv', id="MS", fullName="Proteomics Standards Initiative Mass Spectrometry Ontology", version="4.1.7", URI="https://raw.githubusercontent.com/HUPO-PSI/psi-ms-CV/master/psi-ms.obo")
    ET.SubElement(cvList, 'cv', id="UO", fullName="Unit Ontology", version="14:07:2009", URI="http://obo.cvs.sourceforge.net/*checkout*/obo/obo/ontology/phenotype/unit.obo")
    
    run = ET.SubElement(mzml, 'run', id="run_1", defaultInstrumentConfigurationRef="IC1", startTimeStamp="2023-01-01T00:00:00Z")
    spectrumList = ET.SubElement(run, 'spectrumList', count=str(len(spectra_list)), defaultDataProcessingRef="DP1")
    
    for i, s in enumerate(spectra_list):
        mz_array = s['mz']
        int_array = s['intensity']
        rt = s.get('rt', 0.0)
        scan_id = s.get('id', f"scan={i+1}")
        ms_level = s.get('ms_level', 1)
        
        spectrum = ET.SubElement(spectrumList, 'spectrum', id=scan_id, index=str(i), defaultArrayLength=str(len(mz_array)))
        ET.SubElement(spectrum, 'cvParam', cvRef="MS", accession="MS:1000511", name="ms level", value=str(ms_level))
        ET.SubElement(spectrum, 'cvParam', cvRef="MS", accession="MS:1000579", name="MS1 spectrum", value="")
        
        scanList = ET.SubElement(spectrum, 'scanList', count="1")
        ET.SubElement(scanList, 'cvParam', cvRef="MS", accession="MS:1000795", name="no combination", value="")
        scan = ET.SubElement(scanList, 'scan')
        ET.SubElement(scan, 'cvParam', cvRef="MS", accession="MS:1000016", name="scan start time", value=str(rt), unitCvRef="UO", unitAccession="UO:0000010", unitName="second")
        
        binaryDataArrayList = ET.SubElement(spectrum, 'binaryDataArrayList', count="2")
        
        # M/Z Array
        _add_binary_array(binaryDataArrayList, mz_array, "m/z array", "MS:1000514", "MS:1000040")
        
        # Intensity Array
        _add_binary_array(binaryDataArrayList, int_array, "intensity array", "MS:1000515", "MS:1000040")

    tree = ET.ElementTree(mzml)
    # Pretty print hack
    xmlstr = minidom.parseString(ET.tostring(mzml)).toprettyxml(indent="  ")
    with open(filename, "w") as f:
        f.write(xmlstr)

def _add_binary_array(parent, array, name, accession, unit_accession=None):
    binaryDataArray = ET.SubElement(parent, 'binaryDataArray', encodedLength="0")
    ET.SubElement(binaryDataArray, 'cvParam', cvRef="MS", accession="MS:1000523", name="64-bit float", value="")
    ET.SubElement(binaryDataArray, 'cvParam', cvRef="MS", accession="MS:1000574", name="zlib compression", value="")
    ET.SubElement(binaryDataArray, 'cvParam', cvRef="MS", accession=accession, name=name, value="", unitCvRef="MS" if unit_accession else None, unitAccession=unit_accession, unitName="m/z" if "m/z" in name else "number of counts")
    
    # Encode
    if len(array) > 0:
        data = array.astype(np.float64).tobytes()
        compressed = zlib.compress(data)
        encoded = base64.b64encode(compressed).decode('utf-8')
        
        binary = ET.SubElement(binaryDataArray, 'binary')
        binary.text = encoded
        binaryDataArray.set("encodedLength", str(len(encoded)))
    else:
        binary = ET.SubElement(binaryDataArray, 'binary')
        binary.text = ""




import xml.etree.ElementTree as ET
import base64
import zlib
import numpy as np
import re

class SimpleMzMLReader:
    def __init__(self, filename):
        self.filename = filename
        self.ns = {'mzml': "http://psi.hupo.org/ms/mzml"}

    def _decode_binary(self, binary_element, array_length, cv_params):
        if binary_element.text is None:
            return np.array([])
        
        encoded_data = binary_element.text.strip()
        if not encoded_data:
            return np.array([])
            
        decoded = base64.b64decode(encoded_data)
        
        # Check compression
        is_zlib = 'MS:1000574' in cv_params
        if is_zlib:
            try:
                decoded = zlib.decompress(decoded)
            except Exception as e:
                # If it's supposed to be zlib but fails, that's a critical error
                raise ValueError(f"Zlib decompression failed: {e}")

        # Check precision
        dtype = np.float32
        if 'MS:1000523' in cv_params: # 64-bit float
            dtype = np.float64
        elif 'MS:1000521' in cv_params: # 32-bit float
            dtype = np.float32
        elif 'MS:1000522' in cv_params: # 64-bit integer
            dtype = np.int64
        elif 'MS:1000519' in cv_params: # 32-bit integer
            dtype = np.int32

        return np.frombuffer(decoded, dtype=dtype)

    def get_spectra(self):
        """
        Yields dictionaries containing:
        - rt: retention time in seconds
        - mz: m/z array
        - intensity: intensity array
        - id: spectrum id
        - ms_level: ms level
        """
        context = ET.iterparse(self.filename, events=('end',))
        
        for event, elem in context:
            if elem.tag.endswith('spectrum'):
                spectrum = {}
                spectrum['id'] = elem.attrib.get('id')
                spectrum['index'] = elem.attrib.get('index')
                default_array_length = int(elem.attrib.get('defaultArrayLength', 0))
                
                # Get CV params
                spec_cv_params = elem.findall('mzml:cvParam', self.ns)
                
                # MS Level
                ms_level = 1
                for cv in spec_cv_params:
                    if cv.attrib.get('accession') == 'MS:1000511':
                        ms_level = int(cv.attrib.get('value'))
                        break
                spectrum['ms_level'] = ms_level

                # Scan List for RT
                scan_list = elem.find('mzml:scanList', self.ns)
                rt = 0.0
                if scan_list is not None:
                    scan = scan_list.find('mzml:scan', self.ns)
                    if scan is not None:
                        for cv in scan.findall('mzml:cvParam', self.ns):
                            if cv.attrib.get('accession') == 'MS:1000016':
                                val = float(cv.attrib.get('value'))
                                unit = cv.attrib.get('unitName')
                                if unit == 'minute':
                                    val *= 60
                                rt = val
                                break
                spectrum['rt'] = rt
                
                # Binary Data
                binary_list = elem.find('mzml:binaryDataArrayList', self.ns)
                mz_array = np.array([])
                int_array = np.array([])
                
                if binary_list is not None:
                    for binary_array in binary_list.findall('mzml:binaryDataArray', self.ns):
                        cvs = [x.attrib.get('accession') for x in binary_array.findall('mzml:cvParam', self.ns)]
                        data = self._decode_binary(binary_array.find('mzml:binary', self.ns), default_array_length, cvs)
                        
                        if 'MS:1000514' in cvs: # m/z array
                            mz_array = data
                        elif 'MS:1000515' in cvs: # intensity array
                            int_array = data
                            
                spectrum['mz'] = mz_array
                spectrum['intensity'] = int_array
                
                yield spectrum
                
                # Cleanup to save memory
                elem.clear()

    def get_rts(self):
        """
        Quickly extract all retention times.
        """
        rts = []
        context = ET.iterparse(self.filename, events=('end',))
        for event, elem in context:
            if elem.tag.endswith('spectrum'):
                # Simplified extraction just for RT
                scan_list = elem.find('mzml:scanList', self.ns)
                if scan_list is not None:
                    scan = scan_list.find('mzml:scan', self.ns)
                    if scan is not None:
                        for cv in scan.findall('mzml:cvParam', self.ns):
                            if cv.attrib.get('accession') == 'MS:1000016':
                                val = float(cv.attrib.get('value'))
                                unit = cv.attrib.get('unitName')
                                if unit == 'minute':
                                    val *= 60
                                rts.append(val)
                                break
                elem.clear()
        return np.array(rts)



import re
import numpy as np

def parse_msp(file_path):
    """
    Parses an MSP file and yields spectra dictionaries.
    """
    with open(file_path, 'r', encoding='utf-8', errors='ignore') as f:
        spectrum = {}
        peaks_mz = []
        peaks_intensity = []
        reading_peaks = False
        
        for line in f:
            line = line.strip()
            if not line:
                if spectrum:
                    if peaks_mz:
                        spectrum['mz'] = np.array(peaks_mz, dtype=float)
                        spectrum['intensity'] = np.array(peaks_intensity, dtype=float)
                        yield spectrum
                    spectrum = {}
                    peaks_mz = []
                    peaks_intensity = []
                    reading_peaks = False
                continue
            
            if ':' in line and not reading_peaks:
                key, value = line.split(':', 1)
                key = key.strip()
                value = value.strip()
                
                if key == 'Num Peaks':
                    reading_peaks = True
                    spectrum['num_peaks'] = int(value)
                else:
                    spectrum[key] = value
            elif reading_peaks:
                # Expecting "mz intensity" pair
                parts = line.split()
                if len(parts) >= 2:
                    try:
                        mz = float(parts[0])
                        intensity = float(parts[1])
                        peaks_mz.append(mz)
                        peaks_intensity.append(intensity)
                    except ValueError:
                        pass # Ignore malformed lines

        # Yield the last spectrum if exists
        if spectrum:
            if peaks_mz:
                spectrum['mz'] = np.array(peaks_mz, dtype=float)
                spectrum['intensity'] = np.array(peaks_intensity, dtype=float)
                yield spectrum

def filter_ms1(spectra_generator):
    """
    Filters a generator of spectra for MS1 only.
    """
    for spectrum in spectra_generator:
        if spectrum.get('Spectrum_type') == 'MS1':
            yield spectrum

def load_db(file_path, limit=None):
    """
    Loads database from MSP file, filtering for MS1.
    """
    db = []
    gen = filter_ms1(parse_msp(file_path))
    for i, s in enumerate(gen):
        if limit and i >= limit:
            break
        # Create a simplified object similar to the R list structure
        # name, idms, ionmode, prec, formula, np, rti, instr, msm, spectra(mz, ins)
        
        # Extract fields or use defaults
        # The R code uses 'ins' for notes/comments.
        
        entry = {
            'name': s.get('Name', ''),
            'idms': s.get('DB#', ''),
            'ionmode': s.get('Ion_mode', ''),
            'prec': float(s.get('ExactMass', 0)) if s.get('ExactMass') else 0.0,
            'formula': s.get('Formula', ''),
            'np': s.get('num_peaks', 0),
            'rti': 0, # Placeholder
            'instr': s.get('Instrument_type', ''),
            'msm': s.get('Comments', ''),
            'spectra': {
                'mz': s['mz'],
                'ins': s['intensity']
            }
        }
        db.append(entry)
    return db



import os
import sys
import numpy as np
import pandas as pd
import uuid
import tempfile

# Import from internal mzrtsim module
from .mzrtsim import load_data, simmzml
from .mzrtsim.sim import simmzml_background


def generate_simulation_data(
    n_compounds=100,
    inscutoff=0.05,
    mzrange=(100, 1000),
    rtrange=(0, 600),
    ppm=5,
    sampleppm=5,
    mzdigit=5,
    noise_sd=0.5,
    scanrate=0.1,
    pwidth=10,
    baseline=100,
    baselinesd=30,
    snr=100,
    tailing_factor=1.2,
    compound=None,
    rtime=None,
    tailingindex=None,
    seed=42,
    unique=False,
    matrix=False,
    matrixmz=None,
    output_dir="simulation_output"
):
    """
    Generate simulation data using mzrtsim package.
    
    Args:
        n_compounds (int): Number of compounds to simulate
        inscutoff (float): Intensity cutoff
        mzrange (tuple): m/z range (min, max)
        rtrange (tuple): Retention time range (min, max)
        ppm (float): Mass accuracy in ppm
        sampleppm (float): Sample mass accuracy in ppm
        mzdigit (int): Number of decimal digits for m/z values
        noise_sd (float): Noise standard deviation
        scanrate (float): Scan rate
        pwidth (float): Peak width
        baseline (float): Baseline intensity
        baselinesd (float): Baseline standard deviation
        snr (int): Signal-to-noise ratio
        tailing_factor (float): Peak tailing factor
        compound (list): Specific compounds to use (None for random selection)
        rtime (list): Specific retention times (None for random)
        tailingindex (list): Indices of compounds with tailing (None for all)
        seed (int): Random seed
        unique (bool): Whether to use unique compounds
        matrix (bool): Whether to generate matrix
        matrixmz (list): Matrix m/z values
        output_dir (str): Directory to save output files
        
    Returns:
        dict: Paths to generated files
    """
    # Create output directory if it doesn't exist
    os.makedirs(output_dir, exist_ok=True)
    
    # Load database
    print("Loading database...")
    db = load_data('monahrms1')
    
    # Set all simulation parameters
    params = {
        'n': n_compounds,
        'inscutoff': inscutoff,
        'mzrange': mzrange,
        'rtrange': rtrange,
        'ppm': ppm,
        'sampleppm': sampleppm,
        'mzdigit': mzdigit,
        'noisesd': noise_sd,
        'scanrate': scanrate,
        'pwidth': pwidth,
        'baseline': baseline,
        'baselinesd': baselinesd,
        'SNR': snr,
        'tailingfactor': tailing_factor,
        'compound': compound,
        'rtime': rtime,
        'tailingindex': tailingindex,
        'seed': seed,
        'unique': unique,
        'matrix': matrix,
        'matrixmz': matrixmz
    }
    
    # Generate unique filename without data_type
    filename_prefix = f"simulation_{uuid.uuid4().hex[:8]}"
    
    # Run simulation
    print(f"Generating simulation data...")
    output_path = os.path.join(output_dir, filename_prefix)
    simmzml(db, output_path, **params)
    
    # Prepare output paths
    mzml_path = f"{output_path}.mzML"
    csv_path = f"{output_path}.csv"
    
    return {
        "mzml_path": mzml_path,
        "csv_path": csv_path,
        "params": params
    }


def simulate_background(
    blank_mzml_path,
    data_type="background",
    n_compounds=50,
    noise_sd=0.5,
    snr=100,
    tailing_factor=1.2,
    seed=42,
    output_dir="simulation_output"
):
    """
    Simulate peaks on top of a blank mzML file.
    
    Args:
        blank_mzml_path (str): Path to blank mzML file
        data_type (str): Type of simulation data to generate
        n_compounds (int): Number of compounds to simulate
        noise_sd (float): Noise standard deviation
        snr (int): Signal-to-noise ratio
        tailing_factor (float): Peak tailing factor
        seed (int): Random seed
        output_dir (str): Directory to save output files
        
    Returns:
        dict: Paths to generated files
    """
    # Create output directory if it doesn't exist
    os.makedirs(output_dir, exist_ok=True)
    
    # Load database
    print("Loading database...")
    db = load_data('monahrms1')
    
    # Generate unique filename
    filename_prefix = f"{data_type}_{uuid.uuid4().hex[:8]}"
    output_path = os.path.join(output_dir, filename_prefix)
    
    # Run simulation with background
    print(f"Generating background simulation...")
    simmzml_background(
        blank_mzml_path, db, output_path,
        n=n_compounds,
        noisesd=noise_sd,
        SNR=snr,
        tailingfactor=tailing_factor,
        seed=seed
    )
    
    # Prepare output paths
    mzml_path = f"{output_path}.mzML"
    csv_path = f"{output_path}.csv"
    
    return {
        "mzml_path": mzml_path,
        "csv_path": csv_path,
        "data_type": data_type
    }


if __name__ == "__main__":
    # Example usage
    print("Generating complex high-resolution simulation data...")
    result = generate_simulation_data(
        data_type="complex_hires",
        n_compounds=100,
        output_dir="test_simulation"
    )
    
    print(f"Generated files:")
    print(f"  mzML: {result['mzml_path']}")
    print(f"  CSV: {result['csv_path']}")



import matplotlib.pyplot as plt
import numpy as np
import h5py
from .dataset import DynamicSparseH5Dataset

def plot_sample_image(
    h5_path: str,
    sample_id: str,
    target_rt_precision: float,
    target_mz_precision: float,
    output_path: str = None,
    cmap: str = 'viridis',
    figsize: tuple = (10, 8)
):
    """
    Plots the 2D mass spec image for a given sample ID from an HDF5 file.

    Args:
        h5_path (str): Path to the HDF5 file.
        sample_id (str): The sample ID (e.g., 'Sample_A') to plot.
        target_rt_precision (float): The RT precision to use for image reconstruction.
        target_mz_precision (float): The m/z precision to use for image reconstruction.
        output_path (str, optional): Path to save the plot. If None, displays the plot.
        cmap (str): Colormap for the plot.
        figsize (tuple): Figure size for the plot.
    """
    try:
        dataset = DynamicSparseH5Dataset(
            h5_path=h5_path,
            target_rt_precision=target_rt_precision,
            target_mz_precision=target_mz_precision,
            augment=False # No augmentation for visualization
        )
    except Exception as e:
        print(f"Error loading dataset: {e}")
        return

    # Find the sample index based on sample_id
    sample_idx = -1
    sample_name_found = False
    if 'Sample Name' in dataset.covariates:
        sample_names = [s.decode('utf-8') if isinstance(s, bytes) else s for s in dataset.covariates['Sample Name']]
        if sample_id in sample_names:
            sample_idx = sample_names.index(sample_id)
            sample_name_found = True
    
    if not sample_name_found:
        try:
            # If sample_id is not a name, try to interpret it as an integer index
            sample_idx = int(sample_id)
            if sample_idx < 0 or sample_idx >= len(dataset.filtered_indices):
                raise ValueError("Sample index out of bounds.")
        except ValueError:
            print(f"Error: Sample ID '{sample_id}' not found in 'Sample Name' covariate and is not a valid integer index.")
            return

    # Get the image tensor for the specified sample
    image_tensor, _ = dataset[sample_idx]
    image_np = image_tensor.squeeze().cpu().numpy() # Remove batch and channel dims

    plt.figure(figsize=figsize)
    plt.imshow(image_np, cmap=cmap, aspect='auto',
               extent=[dataset.storage_mz_range[0], dataset.storage_mz_range[1],
                       dataset.storage_rt_range[1], dataset.storage_rt_range[0]]) # RT is usually plotted from top to bottom
    plt.colorbar(label='Intensity')
    plt.xlabel('m/z')
    plt.ylabel('Retention Time (s)')
    plt.title(f'Mass Spec Image for Sample: {sample_id}')
    plt.grid(False)

    if output_path:
        plt.savefig(output_path, bbox_inches='tight')
        print(f"Plot saved to {output_path}")
    else:
        plt.show()

    plt.close()
